#include "bmp.h"
#include "vec2n.h"
#include <assert.h>


//Constructor
BMP::BMP()
{
    m_img = NULL;
}

//Destructor
BMP::~BMP()
{
    if (m_img)
        delete (m_img);
}

unsigned long BMP::GetWidth()
{
    assert (m_img);
	return m_img->GetWidth();
}

unsigned long BMP::GetHeight()
{
    assert (m_img);
    return m_img->GetHeight();
}

bool BMP::IsPixelBlack(Vec2n pos)
{
    assert (m_img);
    assert (! (m_img->GetBPP()%8));
    if (m_img->GetBPP() == 24)
    {
        int width = 4 * (m_img->GetWidth()*3/4+((m_img->GetWidth()*3%4)?1:0));
        int offset = 3*pos.x() + pos.y()*width;
        offset = 4 * (offset/4+((offset%4)?1:0));
        int v1 = m_img->GetImg()[offset+0];
        int v2 = m_img->GetImg()[offset+1];
        int v3 = m_img->GetImg()[offset+2];
        return v1==0 && v2==0 && v3==0;
    }
    else if (m_img->GetBPP() == 8)
    {
        int width = 4 * (m_img->GetWidth()/4+((m_img->GetWidth()%4)?1:0));
        int offset = pos.x()+pos.y()*width;
        int v = m_img->GetImg()[offset];
        unsigned char c1,c2,c3;
        c1 = m_img->GetPalette()[v*3+0];
        c2 = m_img->GetPalette()[v*3+1];
        c3 = m_img->GetPalette()[v*3+2];
        int sum = c1 + c2 + c3;
        return sum ? false : true;
    }
    else
        assert (0);

    return true;
}

void BMP::Clear()
{
	delete (m_img);
}


//Loads bitmap from a BMP file
//Reads headers, dimensions, reads pixel data and stores them
BMPErr BMP::LoadFromFile(std::string fPath)
{
    m_img = new BMPImg;
    BMPErr err = (BMPErr) m_img->Load ((char*) fPath.c_str ());
    if (err != IMG_OK)
    {
        delete m_img;
        m_img = NULL;
        return err;
    }

    return IMG_OK;
}