#pragma once

#include <vector>
#include "errorcharacteristic.h"
#include "rsdecoder.h"

class CodewordDecoder
{
    int GetTotalNumBlocks (const ErrorCharacteristic* error)
    {
        int numBlocks = 0;
        for (int i = 0; i < error->numGroups; i++)
            numBlocks += error->numBlocks[i];
        return numBlocks;
    }
    int GetNumDataCodewords (const ErrorCharacteristic* error)
    {
        int numDataCodewords = 0;
        for (int i = 0; i < error->numGroups; i++)
            numDataCodewords += error->numBlocks[i] * error->numDataCodewordsPerBlock[i];
        return numDataCodewords;
    }

    void GetVecIndices (const ErrorCharacteristic* errchar, std::vector<int>& vecIndices)
    {
        for (int i = 0; i < errchar->numGroups; i++)
            for (int j = 0; j < errchar->numBlocks[i]; j++)
                vecIndices.push_back (i);
    }

    void DivideToBlocks (std::vector<bool> in, std::vector<std::vector<bool>>& out_vecvec, const ErrorCharacteristic* error)
    {
        std::vector<int> vecIndices;
        GetVecIndices (error, vecIndices);

        int numBlocks = GetTotalNumBlocks (error);
        out_vecvec.resize (numBlocks);
        int offset = 0;

        // adding data codewords
        std::vector<int> vecNumCodewords;
        int totalNumCodewords = 0;
        for (int i = 0; i < numBlocks; i++)
        {
            int num = error->numDataCodewordsPerBlock[vecIndices[i]];
            vecNumCodewords.push_back (num);
            totalNumCodewords += num;
        }
        int nSwitch = 0;
        while (totalNumCodewords != 0)
        {
            if (vecNumCodewords[nSwitch] != 0)
            {
                vecNumCodewords[nSwitch]--;
                totalNumCodewords--;

                for (int k = 0; k < 8; k++)
                {
                    out_vecvec[nSwitch].push_back (in[offset]);
                    offset++;
                }
            }
            nSwitch = nSwitch+1 < numBlocks ? nSwitch+1 : 0;
        }

        // adding error codewords
        int numErrorCodewordsPerBlocks = error->numTotalCodewordsPerBlock[0] - error->numDataCodewordsPerBlock[0];
        for (int i = 0; i < numErrorCodewordsPerBlocks; i++)
            for (int j = 0; j < numBlocks; j++)
                for (int numBit = 0; numBit < 8; numBit++)
                {
                    out_vecvec[j].push_back (in[offset]);
                    offset++;
                }
    }

    void SplitData (std::vector<std::vector<bool>> vecSplitted, std::vector<bool>& out_data, 
                const ErrorCharacteristic* error)
    {
        std::vector<int> vecIndices;
        GetVecIndices (error, vecIndices);

        int numBlocks = GetTotalNumBlocks (error);
        out_data.clear ();
        for (int j = 0; j < numBlocks; j++)
            for (int i = 0; i < error->numDataCodewordsPerBlock[vecIndices[j]]; i++)
                for (int numBit = 0; numBit < 8; numBit++)
                    out_data.push_back (vecSplitted[j][i*8+numBit]);
    }


public:
    bool DecodeDataCodewords (std::vector<bool>& in_data, std::vector<bool>& out_data, int version, int errlvl)
    {
        Factory factory;
        StaticData* sd = factory.GetStaticData ();
        const ErrorCharacteristic* error = sd->GetErrorCharacteristic (version, errlvl);
        
        if (((int)in_data.size()) < 8*error->numTotalCodewords)
            throw new ExceptionInvalidParam ("Param in_data.size() is not sufficient big");

        int numDataCodewords = GetNumDataCodewords (error);
        int numBlocks = GetTotalNumBlocks (error);

        std::vector<int> vecIndices;
        GetVecIndices (error, vecIndices);

        std::vector<std::vector<bool>> vecBlocks;
        DivideToBlocks (in_data, vecBlocks, error);

        std::vector<std::vector<bool>> vecCorrectedBlocks (numBlocks);
        RSDecoder decoder;
        for (int i = 0; i < numBlocks; i++)
        {
            decoder.Decode (vecBlocks[i], vecCorrectedBlocks[i], error->numTotalCodewordsPerBlock[vecIndices[i]],
                error->numDataCodewordsPerBlock[vecIndices[i]]);
        }

        SplitData (vecCorrectedBlocks, out_data, error);

        return true;
    }
};