#pragma once

#include "base.h"
#include "errorcharacteristic.h"
#include <fstream>

class ErrorCharacteristicsFactory
{
    int m_prevVersion;
    bool ParseVersion (std::istream& stream, int& out_version)
    {
        int version;
        //std::streampos pos = stream.tellg ();
        stream >> version;
        if (version != m_prevVersion+1)
        {
            //stream.seekg (pos);
            return false;
        }
        out_version = version;

        return true;
    }
    bool ParseInt (std::istream& stream, int& out)
    {
        int value;
        //std::streampos pos = stream.tellg ();
        stream >> value;
        if (! stream.good ())
        {
            //stream.seekg (pos);
            return false;
        }
        out = value;
        return true;
    }
    bool ParseLvl (std::istream& stream, char& out)
    {
        char value;
        std::streampos pos = stream.tellg ();
        stream >> value;
        if (! stream.good ())
        {
            stream.seekg (pos);
            return false;
        }
        switch (value)
        {
        case 'L':
        case 'M':
        case 'Q':
        case 'H':
            break;
        default:
            return false;
        }
        out = value;
        return true;
    }    
#define ERRECCPB {stream.seekg (pos);return false;}
    bool ParseECCPB (std::istream& stream, int& c, int& k, int& r)
    {
        std::streampos pos = stream.tellg ();

        while (' ' == stream.peek () || '\t' == stream.peek() || '\n' == stream.peek())
            stream.get();

        char lb = stream.get ();
        if (lb != '(')
            ERRECCPB;
        stream >> c;
        char comma = stream.get ();
        if (comma != ',')
            ERRECCPB;
        stream >> k;
        comma = stream.get ();
        if (comma != ',')
            ERRECCPB;
        stream >> r;
        comma = stream.get ();
        if (comma != ')')
            ERRECCPB;

        if (! stream.good ())
            ERRECCPB;

        return true;
    }
#undef ERRECCPB

    void Error ()
    {
        throw new ExceptionFatalError ("Data of error characteristics is invalid");
    }

    enum state_e {state_version, state_tnc, state_lvl, state_ecc, state_nb, state_eccpb, state_end};
    bool GetNextStateNB (std::istream& stream, state_e& out)
    {
        std::streampos pos = stream.tellg ();

        int n;
        if (ParseInt (stream, n))
        {
            out = state_nb;
            stream.seekg (pos);
            return true;
        }
        stream.seekg (pos);
        if (ParseECCPB (stream, n, n, n))
        {
            out = state_eccpb;
            stream.seekg (pos);
            return true;
        }
        stream.seekg (pos);
        return false;
    }
    bool GetNextStateECCPB (std::istream& stream, state_e& out)
    {
        std::streampos pos = stream.tellg ();

        int n;
        char c;
        if (ParseVersion (stream, n))
        {
            out = state_version;
            stream.seekg (pos);
            return true;
        }
        stream.seekg (pos);
        if (ParseLvl (stream, c))
        {
            out = state_lvl;
            stream.seekg (pos);
            return true;
        }
        stream.seekg (pos);
        if (ParseECCPB (stream, n, n, n))
        {
            out = state_eccpb;
            stream.seekg (pos);
            return true;
        }

        while (' ' == stream.peek () || '\t' == stream.peek() || '\n' == stream.peek())
            stream.get();

        if (stream.eof ())
        {
            out = state_end;
            stream.seekg (pos);
            return m_prevVersion == 40;
        }
        return false;
    }
    void Parse (std::istream& stream, ErrorCharacteristic* errorChars, int& offset)
    {
        int offsetNB = 0;
        int offsetECCPB = 0;
        int version;
        int tnc;
        char errlvl;


        state_e state = state_version;

        while (true)
        {
            switch (state)
            {
            case state_version:
                if (! ParseVersion (stream, version))
                    Error ();
                m_prevVersion = version;
                state = state_tnc; 
                break;
            case state_tnc:
                if (! ParseInt (stream, tnc))
                    Error ();
                state = state_lvl;
                break;
            case state_lvl:
                if (! ParseLvl (stream, errlvl))
                    Error ();
                state = state_ecc;
                break;
            case state_ecc:
                int necc;
                if (! ParseInt (stream, necc))
                    Error ();
                state = state_nb;
                {
                    ErrorCharacteristic& error = errorChars[offset];
                    error.version = version;
                    error.numTotalCodewords = tnc;
                    error.errorLvl = errlvl;
                    error.numErrCodewords = necc;
                    offset++;
                }
                break;
            case state_nb:
                int nb;
                if (version == 37 && errlvl =='H')
                    printf ("");
                if (! ParseInt (stream, nb))
                    Error ();
                if (! GetNextStateNB (stream, state))
                    Error ();
                {
                    ErrorCharacteristic& error = errorChars[offset-1];
                    error.numBlocks[offsetNB++] = nb;
                    error.numGroups++;
                }
                break;
            case state_eccpb:
                int c,k,r;
                if (! ParseECCPB (stream, c, k, r))
                    Error ();
                if (! GetNextStateECCPB (stream, state))
                    Error ();
                {
                    ErrorCharacteristic& error = errorChars[offset-1];
                    error.numTotalCodewordsPerBlock[offsetECCPB] = c;
                    error.numDataCodewordsPerBlock[offsetECCPB] = k;
                    error.numErrorCorrectionCapPerBlock[offsetECCPB] = r;
                    offsetECCPB++;
                    if (state != state_eccpb)
                    {
                        offsetNB = 0;
                        offsetECCPB = 0;
                        if (offsetNB != offsetECCPB)
                            throw new ExceptionFatalError ("offsetNB != offsetECCPB");
               
                    }
                }

                break;
            case state_end:
                return;
            default:
                Error ();
            }
        }
    }
public:
    ErrorCharacteristicsFactory ()
    {
        m_prevVersion = 0;
    }
    ErrorCharacteristic* Create (CSTR pathToData, ErrorCharacteristic* arrErrorChars)
    {   
        ErrorCharacteristic* error = new ErrorCharacteristic[40*4];
        memset (error, 0, 40*4*sizeof(ErrorCharacteristic));

        int offset = 0;
        std::ifstream file;
        file.open (pathToData);
        if (file.fail ())
        {
            std::string msg = pathToData;
            msg += " cannot be found";
            throw new ExceptionFileNotFound (msg.c_str ());
        }
        Parse (file, error, offset);
        file.close ();

        return error;
    }
};