#pragma once

#include "matrix.h"

class FunctionPatterns
{
private:
    static bool s_isInit;
    static Matrix<bool> s_matPositionPattern;
    static Matrix<bool> s_matPositionPatternTL, s_matPositionPatternTR, s_matPositionPatternBL;
    static Matrix<bool> s_matAlignmentPattern;
    void InitPositionPatternX ()
    {
        s_matPositionPattern.Recreate (7, 7);
        for (int x = 0; x < 7; x++)
            for (int y = 0; y < 7; y++)
                s_matPositionPattern.SetValue (x, y, true);

        for (int x = 1; x < 6; x++)
            for (int y = 1; y < 6; y++)
                s_matPositionPattern.SetValue (x, y, false);

        for (int x = 2; x < 5; x++)
            for (int y = 2; y < 5; y++)
                s_matPositionPattern.SetValue (x, y, true);

        s_matPositionPatternTL.Recreate (8,8);
        for (int x = 0; x < 8; x++)
            for (int y = 0; y < 8; y++)
                s_matPositionPatternTL.SetValue (x, y, false);
        s_matPositionPatternTR = s_matPositionPatternTL.GetDeepCopy ();
        s_matPositionPatternBL = s_matPositionPatternTL.GetDeepCopy ();

        for (int x = 0; x < 7; x++)
            for (int y = 0; y < 7; y++)
                s_matPositionPatternTL.SetValue (x, y, s_matPositionPattern.GetValue (x, y));

        for (int x = 0; x < 7; x++)
            for (int y = 0; y < 7; y++)
                s_matPositionPatternTR.SetValue (x+1, y, s_matPositionPattern.GetValue (x, y));

        for (int x = 0; x < 7; x++)
            for (int y = 0; y < 7; y++)
                s_matPositionPatternBL.SetValue (x, y+1, s_matPositionPattern.GetValue (x, y));
    }

    void InitAlignmentPattern ()
    {
        s_matAlignmentPattern.Recreate (5, 5);
        for (int x = 0; x < 5; x++)
            for (int y = 0; y < 5; y++)
                s_matAlignmentPattern.SetValue (x, y, true);

        for (int x = 1; x < 4; x++)
            for (int y = 1; y < 4; y++)
                s_matAlignmentPattern.SetValue (x, y, false);

        s_matAlignmentPattern.SetValue (2, 2, true);
    }
public:
    // num=0 -> Position pattern
    // num=1 -> Left top position pattern with separators
    // num=2 -> Right top position pattern with separators
    // num=3 -> Left bottom position pattern with separators
    Matrix<bool>& GetPositionPattern (int numVersion = 0)
    {
        switch (numVersion)
        {
        case 0:
            return s_matPositionPattern;
        case 1:
            return s_matPositionPatternTL;
        case 2:
            return s_matPositionPatternTR;
        case 3:
            return s_matPositionPatternBL;
        default:
            throw new ExceptionInvalidParam ("Param numVersion is invalid");
        }   
    }
    Matrix<bool>& GetAlignmentPattern ()
    {
        return s_matAlignmentPattern;
    }
    FunctionPatterns ()
    {
        if (! s_isInit)
        {
            InitPositionPatternX ();
            InitAlignmentPattern ();
            s_isInit = true;
        }
    }
};