#pragma once

#include "matrix.h"
#include "staticdata.h"
#include "factory.h"
#include "functionpatterns.h"
#include "bchdecoder.h"


class QRCode: public Matrix<bool>
{
private:
    
public:
    QRCode ()
    {
    }
    QRCode (Matrix<bool>& mat): Matrix<bool> (mat)
    {
    }

    int GetNumInvalidPositionModulesWithSeparators ()
    {
        FunctionPatterns fp;

        int num = 0;
        Vec2n size = Vec2n(8,8);
        Matrix<bool> mat[3];
        this->GetSubMatrix (mat[0], Vec2n(0,0), size);
        this->GetSubMatrix (mat[1], Vec2n(GetSize().x()-size.x(),0), size);
        this->GetSubMatrix (mat[2], Vec2n(0,GetSize().y()-size.y()), size);
        for (int i=0; i < 3; i++)
            num += mat[i].CompareToMatrix (fp.GetPositionPattern (i+1));
        
        return num;
    }
    int GetNumInvalidPositionModules ()
    {
        FunctionPatterns fp;

        int num = 0;
        Vec2n size = Vec2n(7,7);
        Matrix<bool> mat[3];
        this->GetSubMatrix (mat[0], Vec2n(0,0), size);
        this->GetSubMatrix (mat[1], Vec2n(GetSize().x()-size.x(),0), size);
        this->GetSubMatrix (mat[2], Vec2n(0,GetSize().y()-size.y()), size);
        for (int i=0; i < 3; i++)
            num += mat[i].CompareToMatrix (fp.GetPositionPattern ());
        
        return num;
    } 
    int GetTotalNumPositionModules ()
    {
        return 7*7*3;
    }
    int GetTotalNumPositionModulesWithSeparators ()
    {
        return 8*8*3;
    }

    bool IsNumModulesValid ()
    {
        Factory factory;
        StaticData* psd = factory.GetStaticData ();
        
        if (GetSize().x() != GetSize().y())
            return false;

        return psd->IsCountModulesPerSideValid (GetSize().x());
    }
    bool IsValid ()
    {
        if (! IsNumModulesValid ())
            return false;
        return true;
    }
    void Validate ()
    {
        if (! IsNumModulesValid ())
            throw new ExceptionFatalError ("Size is not valid");
    }
    int GetVersionFromNumModules ()
    {
        Validate ();

        Factory factory;
        StaticData* psd = factory.GetStaticData ();
        return psd->GetVersion (GetSize().x());
    }
    int GetNumModulesPerSide ()
    {
        Validate ();

        return GetSize().x();
    }
    int GetTotalNumAlignmentModules ()
    {
        Validate ();

        Factory factory;
        StaticData* pdata = factory.GetStaticData ();

        int num = 0;
        Vec2n center; 
        for (int i = 0; pdata->GetCenterAlignment (GetVersionFromNumModules(),i,center);i++) 
        {
            num += 5*5;
        }
        return num;
    }

    int GetNumInvalidAlignmentModules ()
    {
        Validate ();

        Factory factory;
        StaticData* pdata = factory.GetStaticData ();

        FunctionPatterns fp;

        int num = 0;
        Vec2n center; 
        for (int i = 0; pdata->GetCenterAlignment (GetVersionFromNumModules(),i,center);i++) 
        {
            Vec2n size (5,5);
            Matrix<bool> submat;
            this->GetSubMatrix (submat, center - Vec2n (2,2), size);

            num += submat.CompareToMatrix (fp.GetAlignmentPattern ());
        }
        return num;
    }

    int GetNumInvalidTimingModules ()
    {
        Validate ();

        Factory factory;
        StaticData* pdata = factory.GetStaticData ();

        bool expectedValue = true;
        int numInvalid = 0;
        for (int x = 8; x < GetSize().x()-8; x++)
        {
            Vec2n v2 (x, 6);
            int num = 0;
            Vec2n center; 
            bool valid = true;
            for (int i = 0; pdata->GetCenterAlignment (GetVersionFromNumModules(),i,center);i++) 
            {
                Vec2n v2min = center - Vec2n (2,2);
                Vec2n v2max = center + Vec2n (2,2);
                if (v2.x() >= v2min.x() && v2.x() <= v2min.x() && v2.y() >= v2min.y() && v2.y() <= v2max.y())
                    valid = false;
            }
            if (valid)
                if (expectedValue != this->GetValue (v2))
                    numInvalid++;
            expectedValue = ! expectedValue;
        }
        expectedValue = true;
        for (int y = 8; y < GetSize().y()-8; y++)
        {
            Vec2n v2 (6, y);
            int num = 0;
            Vec2n center; 
            bool valid = true;
            for (int i = 0; pdata->GetCenterAlignment (GetVersionFromNumModules(),i,center);i++) 
            {
                Vec2n v2min = center - Vec2n (2,2);
                Vec2n v2max = center + Vec2n (2,2);
                if (v2.x() >= v2min.x() && v2.x() <= v2min.x() && v2.y() >= v2min.y() && v2.y() <= v2max.y())
                    valid = false;
            }
            if (valid)
                if (expectedValue != this->GetValue (v2))
                    numInvalid++;
            expectedValue = ! expectedValue;
        }
        return numInvalid;
    }

    int GetTotalNumTimingModules ()
    {
        Validate ();

        Factory factory;
        StaticData* pdata = factory.GetStaticData ();

        int numTotal = 0;
        for (int x = 8; x < GetSize().x()-8; x++)
        {
            Vec2n v2 (x, 6);
            int num = 0;
            Vec2n center; 
            bool valid = true;
            for (int i = 0; pdata->GetCenterAlignment (GetVersionFromNumModules(),i,center);i++) 
            {
                Vec2n v2min = center - Vec2n (2,2);
                Vec2n v2max = center + Vec2n (2,2);
                if (v2.x() >= v2min.x() && v2.x() <= v2min.x() && v2.y() >= v2min.y() && v2.y() <= v2max.y())
                    valid = false;
            }
            if (valid)
                numTotal++;
        }
        for (int y = 8; y < GetSize().y()-8; y++)
        {
            Vec2n v2 (6, y);
            int num = 0;
            Vec2n center; 
            bool valid = true;
            for (int i = 0; pdata->GetCenterAlignment (GetVersionFromNumModules(),i,center);i++) 
            {
                Vec2n v2min = center - Vec2n (2,2);
                Vec2n v2max = center + Vec2n (2,2);
                if (v2.x() >= v2min.x() && v2.x() <= v2min.x() && v2.y() >= v2min.y() && v2.y() <= v2max.y())
                    valid = false;
            }
            if (valid)
                numTotal++;
        }
        return numTotal;
    }

private:
    void GetFormatInfoBitstreamRTLB (bool out_bits[15])
    {
        Validate ();

        int b0to7 = 0;
        for (int i = 0, x=GetSize().x()-1; i < 8; i++,x--)
            b0to7 += GetValue (Vec2n(x,8)) << i;
        
        int b8to14 = 0;
        for (int i = 8, y=GetSize().y()-1-6; i < 15; i++,y++)
            b8to14 += GetValue (Vec2n(8,y)) << i;
        int bitsIn = b0to7 | b8to14;

        memset (out_bits, 0, sizeof (out_bits));
        for (int i = 0; i < 15; i++)
            out_bits[i] = (bitsIn >> (14-i)) & 1;
    }
    void GetFormatInfoBitstreamLT (bool out_bits[15])
    {
        Validate ();

        int b0to5 = 0;
        for (int i = 0; i < 6; i++)
            b0to5 += GetValue(Vec2n(8,i)) << i;
        int b6to7 = 0;
        for (int i = 6; i <= 7 ; i++)
            b0to5 += GetValue(Vec2n(8,i+1)) << i;
        int b8 = GetValue(Vec2n(7, 8)) << 8;
        int b9to14 = 0;
        for (int pos = 0, index=14; pos <= 5; pos++, index--)
            b9to14 += GetValue(Vec2n(pos,8)) << index;

        int bitsIn = b0to5 | b6to7 | b8 | b9to14;

        memset (out_bits, 0, sizeof (out_bits));
        for (int i = 0; i < 15; i++)
            out_bits[i] = (bitsIn >> (14-i)) & 1;
    }

    int GetFormatInfoLT ()
    {
        Validate ();
        bool bitsRaw[15];
        GetFormatInfoBitstreamLT (bitsRaw);

        bool bitsDemasked[15];
        int mask = 0x5412;
        for (int i = 0; i < 15; i++)        
            bitsDemasked[i] = bitsRaw[i] ^ ((mask>>(14-i))&1);
        
        bool bitsDecoded[15];
        BCHDecoder decoder;
        decoder.Decode (bitsDemasked, bitsDecoded, 15, 5);

        int data = 0;
        for (int i = 0; i < 5; i++)
            data |= bitsDecoded[i] << (4-i);
        return data;
    }
    int GetFormatInfoRTLB ()
    {
        Validate ();
        bool bitsRaw[15];
        GetFormatInfoBitstreamRTLB (bitsRaw);

        bool bitsDemasked[15];
        int mask = 0x5412;
        for (int i = 0; i < 15; i++)        
            bitsDemasked[i] = bitsRaw[i] ^ ((mask>>(14-i))&1);
        
        bool bitsDecoded[15];
        BCHDecoder decoder;
        decoder.Decode (bitsDemasked, bitsDecoded, 15, 5);

        int data = 0;
        for (int i = 0; i < 5; i++)
            data |= bitsDecoded[i] << (4-i);
        return data;
    }
public:
    char GetErrorLevelLT ()
    {
        Validate ();

        int format = GetFormatInfoLT ();
        int errlvl = (format >> 3) & 3;
        char clvl;
        switch (errlvl)
        {
        case 1: clvl = 'L'; break;
        case 0: clvl = 'M'; break;
        case 3: clvl = 'Q'; break;
        case 2: clvl = 'H'; break;
        default:
            throw new ExceptionFatalError ("Loaded clvl is invalid");
        }
        return clvl;
    }
    char GetErrorLevelRTLB ()
    {
        Validate ();

        int format = GetFormatInfoRTLB ();
        int errlvl = (format >> 3) & 3;
        char clvl;
        switch (errlvl)
        {
        case 1: clvl = 'L'; break;
        case 0: clvl = 'M'; break;
        case 3: clvl = 'Q'; break;
        case 2: clvl = 'H'; break;
        default:
            throw new ExceptionFatalError ("Loaded clvl is invalid");
        }
        return clvl;
    }

    int GetNumMaskPatternLT ()
    {
        Validate ();

        int format = GetFormatInfoLT ();
        int numMask = (format) & 7;
        return numMask;
    }
    int GetNumMaskPatternRTLB ()
    {
        Validate ();

        int format = GetFormatInfoRTLB ();
        int numMask = (format) & 7;
        return numMask;
    }


private:
    void GetVersionInfoBitstreamLB (bool out_bits[18])
    {
        Validate ();

        if (7 > GetVersionFromNumModules ())
            throw new ExceptionFatalError ("No version info; it is version < 7");
        int index = 0;
        int value = 0;
        for (int x = 0; x < 6; x++)
            for (int y = GetSize().y()-1-7-3; y < GetSize().y()-1-7; y++)
            {
                value += GetValue (Vec2n(x,y)) << index;
                index++;
            }
        for (int i = 0; i < 18; i++)
            out_bits[i] = value>>(17-i) & 1;
    }
    void GetVersionInfoBitstreamRT (bool out_bits[18])
    {
        Validate ();

        if (7 > GetVersionFromNumModules ())
            throw new ExceptionFatalError ("No version info; it is version < 7");
        int index = 0;
        int value = 0;
        for (int y = 0; y < 6; y++)
            for (int x = GetSize().x()-1-7-3; x < GetSize().x()-1-7; x++)
            {
                value += GetValue (Vec2n(x,y)) << index;
                index++;
            }
        for (int i = 0; i < 18; i++)
            out_bits[i] = value>>(17-i) & 1;
    }
public:
    int GetVersionFromVersionInfoLB ()
    {
        Validate ();

        int version;
        bool bitsRaw [18];
        GetVersionInfoBitstreamLB (bitsRaw);

        bool bitsDecoded [18];
        BCHDecoder decoder;
        decoder.Decode (bitsRaw, bitsDecoded, 18, 6);
        for (int i = 0; i < 6; i++)
            version |= bitsDecoded[i] << (5-i);
        return version;
    }
    int GetVersionFromVersionInfoRT ()
    {
        Validate ();

        int version;
        bool bitsRaw [18];
        GetVersionInfoBitstreamRT (bitsRaw);

        bool bitsDecoded [18];
        BCHDecoder decoder;
        decoder.Decode (bitsRaw, bitsDecoded, 18, 6);
        for (int i = 0; i < 6; i++)
            version |= bitsDecoded[i] << (5-i);
        return version;
    }
    bool IsVersionInfo ()
    {
        Validate ();

        return (7 <= GetVersionFromNumModules ());
    }

    bool IsPositionModule (int x, int y)
    {
        Validate ();

        // 1st region:
        if (x >= 0 && y >= 0 && x <= 6 && y <= 6)
            return true;

        int maxX = GetSize().x()-1;
        int maxY = GetSize().y()-1;
        
        // 2nd region:
        if (x >= maxX-6 && y >= 0 && x <= maxX && y <= 6)
            return true;

        if (y >= maxY-6 && x >= 0 && y <= maxY && x <= 6)
            return true;

        return false;
    }

    bool IsPositionSeparator (int x, int y)
    {
        Validate ();

        if (IsPositionModule (x, y))
            return false;

        // 1st region:
        if (x >= 0 && y >= 0 && x <= 7 && y <= 7)
            return true;

        int maxX = GetSize().x()-1;
        int maxY = GetSize().y()-1;
        
        // 2nd region:
        if (x >= maxX-7 && y >= 0 && x <= maxX && y <= 7)
            return true;

        if (y >= maxY-7 && x >= 0 && y <= maxY && x <= 7)
            return true;

        return false;        
    }

    bool IsAlignmentModule (int x, int y)
    {
        Validate ();
        Factory factory;
        StaticData* sd = factory.GetStaticData ();

        Vec2n center;
        for (int i = 0; sd->GetCenterAlignment (GetVersionFromNumModules (), i, center); i++)
        {
            if (x <= center.x()+2 && x >= center.x()-2 && y <= center.y()+2 && y >= center.y()-2)
                return true;
        }
        return false;
    }

    bool IsTimingModule (int x, int y)
    {
        Validate ();

        if (IsPositionModule (x, y))
            return false;
        if (IsAlignmentModule (x, y))
            return false;

        if (x == 6 || y == 6)
            return true;

        return false;

    }

    bool IsVersionInfoModule (int x, int y)
    {
        Validate ();

        if (7 > GetVersionFromNumModules ())
            return false;
        
        if (x >= GetSize().x()-1-7-3 && x <= GetSize().x()-1-7-1 && y < 6)
            return true;

        if (y >= GetSize().y()-1-7-3 && y <= GetSize().y()-1-7-1 && x < 6)
            return true;

        return false;
    }

    bool IsFormatInfoModule (int x, int y)
    {
        Validate ();

        if (IsTimingModule (x, y))
            return false;
        if (x == 8 && y <= 8)
            return true;
        if (x <= 8 && y == 8)
            return true;

        int maxX = GetSize().x()-1;
        int maxY = GetSize().y()-1;

        if (x == 8 && y >= maxY-7)
            return true;
        if (y == 8 && x >= maxX-7)
            return true;
        return false;
    }

    bool IsCodewordModule (int x, int y)
    {
        Validate ();

        if (IsPositionModule (x, y) ||
                    IsPositionSeparator (x, y) ||
                    IsTimingModule (x, y) ||
                    IsAlignmentModule (x, y) ||
                    IsFormatInfoModule (x, y) ||
                    IsVersionInfoModule (x, y) ||
                    false)
            return false;

        return true;
    }

    void PrintOrderOfCodewords ();

    void PrintOrderOfBits ();

    void PrintBitsDemasked (int mask)
    {
        Factory factory;
        StaticData* sd = factory.GetStaticData ();
        for (int y = 0; y<GetSize().y(); y++)
        {
            for (int x = 0; x<GetSize().x(); x++)
            {
                bool value = GetValue(Vec2n(x,y));
                value ^= sd->GetMaskAtPos (x, y, mask);
                putchar (value ? 'X' : '.');
            }
            printf ("\n");
        }
    }
};