#pragma once

#include <vector>
#include "vec2n.h"
#include "base.h"
#include "exception.h"
#include "errorcharacteristicfactory.h"
#include "errorcharacteristic.h"

class Factory;

#pragma warning (disable : 4996)
class StaticData
{
    friend Factory;
    bool m_isInitialized;
    void CheckIfInit ()
    {
        if (! m_isInitialized)
            throw new ExceptionFatalError ("StaticData is not initialized");
    }

    std::vector <std::vector<int>> m_vecvecAlignCenters;
    void InitAlignCenters (CSTR pathToFile)
    {
        m_vecvecAlignCenters.clear ();
        m_vecvecAlignCenters.resize(41);

        FILE* file = fopen (pathToFile, "r");
        if (! file)
        {
            std::string msg = pathToFile;
            msg += " cannot be found";
            throw new ExceptionFileNotFound (msg.c_str ());
        }
        int count = 0;
        for (int i = 0; i <= 40; i++)
        {
            if (i == 2)
                count = 2;
            if (i == 7)
                count = 3;
            if (i == 14)
                count = 4;
            if (i == 21)
                count = 5;
            if (i == 28)
                count = 6;
            if (i == 35)
                count = 7;

            for (int j = 0; j < count; j++)
            {
                int value;
                fscanf (file, "%d", &value);
                m_vecvecAlignCenters[i].push_back (value);
            }
        }
        fclose (file);
    }
    // private! Use Factory!
    StaticData (CSTR pathToData)
    {
        m_isInitialized = false;
        m_arrErrorChars = NULL;

        std::string pathToAlignCenters (pathToData);
        pathToAlignCenters += "\\centersOfAlignPatterns.dat";
        InitAlignCenters (pathToAlignCenters.c_str());

        std::string pathToErrorChars (pathToData);
        pathToErrorChars += "\\errorCharacteristics.dat";      
        InitErrorCharacteristic (pathToErrorChars.c_str());

        m_isInitialized = true;
    }
    ~StaticData ()
    {
        SAFE_DELETE_ARRAY (m_arrErrorChars);
    }
public:
    static bool IsVersionValid (int version)
    {
        return version > 0 && version <= 40;
    }
    static int GetModulesPerSide (int version)
    {
        if (! IsVersionValid (version))
            throw new ExceptionInvalidParam ("Param version is invalid");
        return 21+(version-1)*4;
    }

    static bool IsErrorLevelValid (unsigned char errlvl)
    {
        switch (errlvl)
        {
        case 'L':
        case 'Q':
        case 'M':
        case 'H':
            return true;
        default:
            return false;
        }
    }

    static bool IsCountModulesPerSideValid (int modulesPerSide)
    {
        return ! (modulesPerSide < 21 || modulesPerSide > 177 || ((modulesPerSide-21)%4!=0));
    }
    static int GetVersion (int modulesPerSide)
    {
        if (! IsCountModulesPerSideValid (modulesPerSide))
            throw new ExceptionInvalidParam ("Param modulesPerSide is invalid");
        return (modulesPerSide-21)/4+1;
    }

    // use: 
    // Vec2n center; 
    // for (int i = 0; GetCenterAlignment (version,i,center);i++) {...}
    bool GetCenterAlignment (int version, int num, Vec2n& center)
    {
        CheckIfInit ();

        if (! IsVersionValid (version))
            throw new ExceptionInvalidParam ("Param version is invalid");
        std::vector<int>& vec = m_vecvecAlignCenters[version];
        if ((int)(vec.size()*vec.size()-3) <= num)
            return false;
        
        // alignment pattern is not placed everywhere... Some places is skipped:
        num++;
        if (num >= (int)(vec.size()-1))
            num++;
        if (num >= (int)(vec.size()*(vec.size()-1)))
            num++;
        
        int x = vec[num%vec.size()];
        int y = vec[num/vec.size()];
        center.set (x, y);
        
        return true;
    }


    static bool GetMaskAtPos (int x, int y, int mask)
    {
        int i = y;
        int j = x;
        switch (mask)
        {
        case 0:
            return ! ((i+j)%2);
        case 1:
            return ! (i%2);
        case 2:
            return ! (j%3);
        case 3:
            return ! ((i+j) % 3);
        case 4:
            return ! (((i/2)+(j/3)) % 2);
        case 5:
            return ! ((i*j)%2+(i*j)%3);
        case 6:
            return ! (((i*j)%2+(i*j)%3)%2);
        case 7:
            return ! (((i*j)%3+(i+j)%2)%2);
        default:
            throw new ExceptionInvalidParam ("Param mask is invalid");
        }
        return 0;
    }

private:
    // 160 items:
    ErrorCharacteristic* m_arrErrorChars;
    void InitErrorCharacteristic (CSTR pathToFile)
    {
        ErrorCharacteristicsFactory factory;
        m_arrErrorChars = factory.Create (pathToFile, m_arrErrorChars);
    }

public:
    const ErrorCharacteristic* GetErrorCharacteristic (int version, char errorlevel)
    {
        CheckIfInit ();

        if (! IsVersionValid (version))
            throw new ExceptionInvalidParam ("Param version is invalid");
        
        int index = (version-1)*4;
        switch (errorlevel)
        {
        case 'H': index++;
        case 'Q': index++;
        case 'M': index++;
        case 'L': 
            break;
        default:
            throw new ExceptionInvalidParam ("Param errorlevel is invalid");
        }
        ErrorCharacteristic* error = &(m_arrErrorChars[index]);
        if (error->errorLvl != errorlevel || error->version != version)
            throw new ExceptionFatalError ("Fatal error");
        return error;
    }

};