#pragma once

/** Dvojica int-ov pre vseobecne pouzitie. */
class Vec2n
{
public:
    /** Typ premennych, ktora trieda obaluje.*/
    typedef int value_type;


    /** Pocet komponent v triede. */
    enum { num_components = 2 };
    

    /** Pole o 2 prvkoch */
    value_type _v[2];


    /** Konstruktor vynuluje objekt. */
    Vec2n() 
    {
        _v[0]=0;
        _v[1]=0;
    }
        

    /** Konstruktor inicializuje trojicu konretnymi hodnotami.
     \param u hodnota 0.-teho prvku
     \param v hodnota 1.-teho prvku*/
    Vec2n(value_type u, value_type v) 
    {
        _v[0]=u;
        _v[1]=v;
    }


    //** Operator skonvertuje \a Vec2n na trojicu floatov */
    //inline operator osg::Vec2f () const 
    //{
    //    return osg::Vec2f ((float) x(), (float) y()); 
    //}
   

    /** Porovna s inou dvojicou
     \param v porovnavany vektor
     \retval vrati vysledok porovnavania. Ak su si dvojice equivalentne vrati true, inac false.*/
    inline bool operator == (const Vec2n& v) const 
    { 
        return _v[0]==v._v[0] && _v[1]==v._v[1];
    }


    /** Porovna s inou dvojicou
     \param v porovnavany vektor
     \retval vrati vysledok porovnavania. Ak nie su dvojice equivalentne vrati true, inac false.*/
    inline bool operator != (const Vec2n& v) const 
    {
        return _v[0]!=v._v[0] || _v[1]!=v._v[1];
    }


    /** Porovna s inou dvojicou. Najprv porovna 0.te prvky, ak su si rovne, vrati vysledok porovnania prvkov na indexe 1.
     \param v porovnavany vektor
     \retval vrati vysledok porovnavania.*/
    inline bool operator <  (const Vec2n& v) const
    {
        if (_v[0] < v._v[0]) 
            return true;
        else if (_v[0] > v._v[0]) 
            return false;
        else 
            return (_v[1] < v._v[1]);
    }


    /** Pretypuje pole na ukazatel na int.
     \retval vrati ukazatel na pamat, kde sa nachadza pole \a _v*/ 
    inline value_type* ptr() 
    {
        return _v;
    }


    /** Pretypuje pole na ukazatel na konstany int.
     \retval vrati ukazatel na pamat, kde sa nachadza pole \a _v*/
    inline const value_type* ptr() const
    { 
        return _v;
    }
    

    /** Metoda nastavuje nove hodnoty dvojice.
     \param u nova hodnota 0.-teho prvku.
     \param v nova hodnota 1.-eho prvku.
     */
    inline void set(value_type u, value_type v)
    {
        _v[0]=u; _v[1]=v;
    }


    /** Metoda nastavuje nove hodnoty dvojici.
     \param rhs nove hodnoty dvojice
     */
    inline void set (const Vec2n& rhs)
    {
        _v[0]=rhs._v[0]; _v[1]=rhs._v[1];
    }


    /** Vrati komponentu na indexe \a i.
     \param i index komponenty
     \retval vrati hodnotu komponenty
     */
    inline value_type& operator [] (unsigned int i) 
    {
        return _v[i];
    }


    /** Vrati komponentu na indexe \a i.
     \param i index komponenty
     \retval vrati hodnotu komponenty
     */
    inline value_type operator [] (unsigned int i) const
    {
        return _v[i];
    }


    /** Vrati komponentu na indexe 0.
     \retval vrati hodnotu komponenty
     */
    inline value_type& x() 
    {
        return _v[0];
    }


    /** Vrati komponentu na indexe 1.
     \retval vrati hodnotu komponenty
     */
    inline value_type& y() 
    {
        return _v[1];
    }


    /** Vrati komponentu na indexe 0.
     \retval vrati hodnotu komponenty
     */
    inline value_type x() const 
    {
        return _v[0];
    }

    
    /** Vrati komponentu na indexe 1.
     \retval vrati hodnotu komponenty
     */
    inline value_type y() const
    {
        return _v[1];
    }


    /** Vrati komponentu na indexe 0.
     \retval vrati hodnotu komponenty
     */
    inline value_type& u()
    {
        return _v[0];
    }


    /** Vrati komponentu na indexe 1.
     \retval vrati hodnotu komponenty
     */
    inline value_type& v() 
    {
        return _v[1];
    }


    /** Vrati komponentu na indexe 0.
     \retval vrati hodnotu komponenty
     */
    inline value_type u() const 
    { 
        return _v[0]; 
    }


    /** Vrati komponentu na indexe 1.
     \retval vrati hodnotu komponenty
     */
    inline value_type v() const 
    { 
        return _v[1]; 
    }


    /** Binarne nasobenie vektoru so skalarom. 
     \param rhs skalar
     \retval vrati vysledok nasobenia*/
    inline Vec2n operator * (value_type rhs) const
    {
        Vec2n col(*this);
        col *= rhs;
        return col;
    }


    /** Unarne nasobenie vektor skalarom.
     \param rhs skalar
     \retval vrati vysledok nasobenia */
    inline Vec2n& operator *= (value_type rhs)
    {
        _v[0] = (value_type)(_v[0]*rhs);
        _v[1] = (value_type)(_v[1]*rhs);
        return *this;
    }


    /** Binarne delenie vektora skalarom.
     \param rhs skalar
     \retval vrati vysledok delenia */
    inline Vec2n operator / (value_type rhs) const
    {
        Vec2n col(*this);
        col /= rhs;
        return col;
    }


    /** Unarne delenie vektora skalarom.
     \param rhs skalar
     \retval vrati vysledok delenia */
    inline Vec2n& operator /= (value_type rhs)
    {
        this->set (x()/rhs, y()/rhs);
        return *this;
    }


    /** Binarne scitavanie vektora vektorom.
     \param rhs vektor
     \retval vrati vysledok scitavania */
    inline Vec2n operator + (const Vec2n& rhs) const
    {
        return Vec2n (_v[0]+rhs._v[0], _v[1]+rhs._v[1]);
    }


    /** Unarne scitavanie vektora vektorom.
     \param rhs vektor
     \retval vrati vysledok scitavania */
    inline Vec2n& operator += (const Vec2n& rhs)
    {
        _v[0] += rhs._v[0];
        _v[1] += rhs._v[1];
        return *this;
    }


    /** Binarne odcitanie vektora vektorom.
     \param rhs vektor
     \retval vrati vysledok odcitavania */
    inline Vec2n operator - (const Vec2n& rhs) const
    {
        return Vec2n (_v[0]-rhs._v[0], _v[1]-rhs._v[1]);
    }


    /** Unarne odcitanie vektorov.
     \param rhs vektor
     \retval vrati vysledok odcitavania */
    inline Vec2n& operator -= (const Vec2n& rhs)
    {
        _v[0] -= rhs._v[0];
        _v[1] -= rhs._v[1];
        return *this;
    }


    /** Metoda porovna dva \a Vec2n a vrati \a Vec2n, ktory na 0. pozicii bude mat minimum z v1[0] a v2[0], na 1. pozicii minimum z v1[1] a v2[1].
     \param v1 1. dvojica
     \param v2 2. dvojica
     \retval vrati minimum z dvojic*/
    static Vec2n min (const Vec2n& v1, const Vec2n& v2)
    {
        return Vec2n (
            std::min (v1.x (), v2.x ()),
            std::min (v1.y (), v2.y ()));
    }


    /** Metoda porovna dva \a Vec2n a vrati \a Vec2n, ktory na 0. pozicii bude mat maximum z v1[0] a v2[0], na 1. pozicii maximum z v1[1] a v2[1].
     \param v1 1. dvojica
     \param v2 2. dvojica
     \retval vrati maximum z dvojic*/    
    static Vec2n max (const Vec2n& v1, const Vec2n& v2)
    {
        return Vec2n (
            std::max (v1.x (), v2.x ()),
            std::max (v1.y (), v2.y ()));
    }
};